SystemOrganization addCategory: #'Zinc-Server-Sent-Events'!

Object subclass: #ZnServerSentEvent
	instanceVariableNames: 'data event id retry'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Zinc-Server-Sent-Events'!

!ZnServerSentEvent commentStamp: '<historical>' prior: 0!
A ZnServerSentEvent is used to send data from a server to an HTTP client, on the initiative of the server, often called a push notification.

ZnServerEvents are written out according to the text/event-stream MIME type.

Instances variables

 data <String> the actual data without LF, CR or CRLF
 event <String> optional event type tag
 id <String> optional event id tag
 retry <Integer> optional number of milliseconds for connection retry

References

 http://en.wikipedia.org/wiki/Server-sent_events
 http://dev.w3.org/html5/eventsource/
 https://developer.mozilla.org/en-US/docs/Server-sent_events/Using_server-sent_events

  ZnServerSentEvent type: 'counter' data: 100.

Part of Zinc HTTP Components!

!ZnServerSentEvent class methodsFor: 'instance creation' stamp: 'SvenVanCaekenberghe 8/19/2013 12:36'!
data: string
	^ self new
		data: string;
		yourself! !

!ZnServerSentEvent class methodsFor: 'instance creation' stamp: 'SvenVanCaekenberghe 8/19/2013 12:37'!
id: id data: string
	^ self new
		id: id;
		data: string;
		yourself! !

!ZnServerSentEvent class methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 8/19/2013 11:47'!
mimeType
	^ 'text/event-stream' asZnMimeType! !

!ZnServerSentEvent class methodsFor: 'instance creation' stamp: 'SvenVanCaekenberghe 8/19/2013 12:36'!
type: tag data: string
	^ self new
		event: tag;
		data: string;
		yourself! !

!ZnServerSentEvent class methodsFor: 'instance creation' stamp: 'SvenVanCaekenberghe 8/19/2013 12:37'!
type: tag id: id data: string
	^ self new
		event: tag;
		id: id;
		data: string;
		yourself! !

!ZnServerSentEvent methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 8/19/2013 11:46'!
data
	^ data! !

!ZnServerSentEvent methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 8/19/2013 12:04'!
data: string
	"string cannot contain CR, LF or CRLF"
	
	data := string! !

!ZnServerSentEvent methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 8/19/2013 11:46'!
event
	^ event! !

!ZnServerSentEvent methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 8/19/2013 12:04'!
event: typeString
	event := typeString! !

!ZnServerSentEvent methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 8/19/2013 11:46'!
id
	^ id! !

!ZnServerSentEvent methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 8/19/2013 12:04'!
id: idString
	id := idString! !

!ZnServerSentEvent methodsFor: 'printing' stamp: 'SvenVanCaekenberghe 8/26/2013 15:18'!
printOn: stream
	| first |
	super printOn: stream.
	stream nextPut: $( .
	first := true.
	#( 'event' 'id' 'retry' 'data' ) 
		do: [ :field |
			(self instVarNamed: field) 
				ifNotNil: [ :value | 
					first ifTrue: [ first := false ] ifFalse: [ stream << ', '  ].
					stream << field; nextPut: $= ; << value ] ].
	stream nextPut: $)! !

!ZnServerSentEvent methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 8/19/2013 11:46'!
retry
	^ retry! !

!ZnServerSentEvent methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 8/19/2013 12:04'!
retry: milliseconds
	retry := milliseconds! !

!ZnServerSentEvent methodsFor: 'writing' stamp: 'SvenVanCaekenberghe 8/26/2013 15:30'!
writeOn: stream
	| encoder  |
	encoder := ZnCharacterWriteStream on: stream.
	#( 'data' 'event' 'id' 'retry' ) do: [ :field |
		(self instVarNamed: field) 
			ifNotNil: [ :value | 
				encoder << field; nextPut: $: ; << value; << String crlf ] ].
	encoder << String crlf; flush.
	ZnCurrentServer value 
		ifNotNil: [ :server | server log info: [ 'Wrote ', self printString ] ]! !

Object subclass: #ZnServerSentEventDelegate
	instanceVariableNames: 'prefix handler'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Zinc-Server-Sent-Events'!

!ZnServerSentEventDelegate commentStamp: '<historical>' prior: 0!
ZnServerSentEventDelegate is an example of a Server-Sent Events application.

My #clientHtml listens to simple status update event sent by the server, using ZnServerSentEventStatusHandler.

Installation

  ZnServerSentEventDelegate installInServer: ZnServer default.

Part of Zinc HTTP Components.!

!ZnServerSentEventDelegate class methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 8/26/2013 15:29'!
clientHtml
	^ '<html>
	<head>
	<title>Zinc Server-Side Events Status Demo</title>
	<script language="javascript" type="text/javascript">
	var source, statusElement;
	function init() {
		statusElement = document.getElementById("status");
		var source = new EventSource(''sse-status'');
		source.addEventListener(
			''status-update'', 
			function(event) {
				setStatus(event.data); 
			},
			false);
	}
	function setStatus(text) {
		statusElement.innerHTML = text;
	}
	window.addEventListener("load", init, false);
	</script>
	</head>
	<body>
	<h1>Zinc Server-Side Events Status Demo</h1>
	<div><pre id="status"></pre></div>
	</body>
	</html>'! !

!ZnServerSentEventDelegate class methodsFor: 'convenience' stamp: 'SvenVanCaekenberghe 8/26/2013 11:35'!
installInServer: server
	server delegate
		map: 'sse-status-client'
		to: [ :request | ZnResponse ok: (ZnEntity html: self clientHtml) ];
		map: 'sse-status'
		to: self new! !

!ZnServerSentEventDelegate methodsFor: 'public' stamp: 'SvenVanCaekenberghe 8/26/2013 16:16'!
handleRequest: request
	(prefix isNil or: [ prefix = request uri pathSegments ])
		ifFalse: [ ^ ZnResponse notFound: request uri ].
	^ self serverSentEventResponse! !

!ZnServerSentEventDelegate methodsFor: 'initialize-release' stamp: 'SvenVanCaekenberghe 8/26/2013 16:16'!
initialize
	super initialize.
	prefix := 'sse-status'.
	handler := ZnServerSentEventStatusHandler new! !

!ZnServerSentEventDelegate methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 8/26/2013 16:17'!
serverSentEventResponse
	| response |
	response := ZnServerSentEventResponse new.
	response
		statusLine: ZnStatusLine ok;
		headers: ZnHeaders defaultResponseHeaders;
		handler: handler.
	response headers
		at: 'Content-Type' put: ZnServerSentEvent mimeType printString.
	^ response! !

!ZnServerSentEventDelegate methodsFor: 'public' stamp: 'SvenVanCaekenberghe 8/26/2013 11:39'!
value: request
	^ self handleRequest: request! !

Object subclass: #ZnServerSentEventStatusHandler
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Zinc-Server-Sent-Events'!

!ZnServerSentEventStatusHandler commentStamp: '<historical>' prior: 0!
ZnServerSentEventStatusHandler is an example of a Server Sent Events application.

This handler will send status updates every 2 seconds.

Part of Zinc HTTP Components.!

!ZnServerSentEventStatusHandler methodsFor: 'running' stamp: 'SvenVanCaekenberghe 8/26/2013 15:42'!
run: connection
	| event status |
	[ 
	status := self statusString.
	event := ZnServerSentEvent type: 'status-update' data: status.
	event writeOn: connection.
	2 seconds asDelay wait ] repeat! !

!ZnServerSentEventStatusHandler methodsFor: 'acccessing' stamp: 'SvenVanCaekenberghe 8/26/2013 15:51'!
statusString
	| vm |
	vm := SmalltalkImage current vm.
	^ String streamContents: [ :out | 
		out print: DateAndTime now;
		space;
		<< 'memory '; print: vm memoryEnd; << ' bytes ' ;
		print: (vm memoryEnd - vm youngSpaceEnd / vm memoryEnd * 100.0); << '% free' ]! !

ZnResponse subclass: #ZnServerSentEventResponse
	instanceVariableNames: 'handler'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Zinc-Server-Sent-Events'!

!ZnServerSentEventResponse commentStamp: '<historical>' prior: 0!
A ZnServerSentEventResponse is a ZnResponse that will continue using the open HTTP connection by passing it to handler's #run: method.

Part of Zinc HTTP Components.
!

!ZnServerSentEventResponse methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 8/26/2013 14:51'!
handler: anObject
	handler := anObject! !

!ZnServerSentEventResponse methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 8/26/2013 15:52'!
useConnection: connection
	"Hook method overwritten to give the receiver the chance to 
	keep using connection in the current thread/process after the server wrote the response.
	We pass control to our handler."
	
	[ handler run: connection ]
		on: ConnectionClosed, ConnectionTimedOut
		do: [ "ignore" ]! !

!ZnServerSentEventResponse methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 8/26/2013 14:51'!
wantsConnectionClose
	"After we are done, the server can close the connection we were using"
	 
	^ true! !
